#include <valarray>
#include <vector>
#include <istream>
#include <ostream>
#include <cstddef>
#include <algorithm>
#include <cmath>
#include <stdint.h>
#include "arith.h"
#include "coders2d.h"

/****************
 * Utility code *
 ****************/
static inline double length(const std::valarray<int> &x)
{
    return sqrt((double) x[0] * (double) x[0]
                + (double) x[1] * (double) x[1]);
}

static inline int round_length(const std::valarray<int> &x)
{
    return (int) round(length(x));
}

/************************************************************
 * Axial coder: X, Y and Z are encoded in separate contexts *
 ************************************************************/

axial_coder::axial_coder(int max_x, int max_y)
: max_x(max_x), max_y(max_y)
{
    axes[0].code_range(-max_x, max_x);
    axes[1].code_range(-max_y, max_y);
}

axial_coder::axial_coder(std::istream &in)
{
    uint32_t mx, my;
    in.read((char *) &mx, sizeof(mx));
    in.read((char *) &my, sizeof(my));
    max_x = mx; max_y = my;
    axes[0].code_range(-max_x, max_x);
    axes[1].code_range(-max_y, max_y);
}

double axial_coder::encode(compressor &c, const std::valarray<int> &d)
{
    double b = 0.0;
    b += c.encode(axes[0], d[0]);
    b += c.encode(axes[1], d[1]);
    return b;
}

std::valarray<int> axial_coder::decode(decompressor &d)
{
    std::valarray<int> ans(2);
    ans[0] = d.decode(axes[0]);
    ans[1] = d.decode(axes[1]);
    return ans;
}

double axial_coder::bits() const
{
    return axes[0].bits() + axes[1].bits();
}

double axial_coder::entropy() const
{
    return axes[0].entropy() + axes[1].entropy();
}

std::size_t axial_coder::use() const
{
    return axes[0].use();
}

void axial_coder::dump(std::ostream &out) const
{
    out << "X:\n" << axes[0] << "Y:\n" << axes[1];
}

void axial_coder::write_header(std::ostream &out) const
{
    uint32_t mx = max_x;
    uint32_t my = max_y;
    out.write((const char *) &mx, sizeof(mx));
    out.write((const char *) &my, sizeof(my));
}

/*******************************************************************
 * Single context coder: similar to axial_coder, but uses only one *
 * context for coordinates.                                        *
 *******************************************************************/

single_coder::single_coder(int max_x, int max_y)
{
    max_xy = std::max(max_x, max_y);
    freq.code_range(-max_xy, max_xy);
}

single_coder::single_coder(std::istream &in)
{
    uint32_t m;
    in.read((char *) &m, sizeof(m));
    max_xy = m;
    freq.code_range(-m, m);
}

double single_coder::encode(compressor &c, const std::valarray<int> &d)
{
    double b = 0.0;
    b += c.encode(freq, d[0]);
    b += c.encode(freq, d[1]);
    return b;
}

std::valarray<int> single_coder::decode(decompressor &d)
{
    std::valarray<int> ans(2);
    ans[0] = d.decode(freq);
    ans[1] = d.decode(freq);
    return ans;
}

double single_coder::bits() const
{
    return freq.bits();
}

double single_coder::entropy() const
{
    return freq.entropy();
}

std::size_t single_coder::use() const
{
    return freq.use() / 2;
}

void single_coder::dump(std::ostream &out) const
{
    out << "freq:\n" << freq;
}

void single_coder::write_header(std::ostream &out) const
{
    uint32_t m = max_xy;
    out.write((const char *) &m, sizeof(m));
}

/***************************************************************
 * Radial coder: encodes (int) radius, then index with uniform *
 ***************************************************************/

radial_coder::radial_coder(int max_length)
:
max_length(max_length),
rfreq(),
sizes(),
index(),
total_bits(0.0), radial_entropy(0.0)
{
    initialise(true);
}

radial_coder::radial_coder(std::istream &in)
:
max_length(0),
rfreq(),
sizes(),
index(),
total_bits(0.0), radial_entropy(0.0)
{
    uint32_t ml;

    in.read((char *) &ml, sizeof(ml));
    max_length = ml;
    initialise(false);
}

void radial_coder::initialise(bool compressing)
{
    sizes.resize(max_length + 1, 0);
    if (compressing)
        index.resize(2 * max_length + 1, std::vector<int>(2 * max_length + 1, 0));
    else
        rindex.resize(max_length + 1);
    rfreq.code_range(0, max_length);

    std::valarray<int> xy(2);
    for (int y = -max_length; y <= max_length; y++)
        for (int x = -max_length; x <= max_length; x++)
        {
            xy[0] = x;
            xy[1] = y;
            int l = round_length(xy);
            if (l <= max_length)
            {
                radial_address v;
                v.x = x; v.y = y;
                if (compressing)
                    index[y + max_length][x + max_length] = sizes[l];
                else
                    rindex[l].push_back(v);
                sizes[l]++;
            }
        }
}

double radial_coder::encode(compressor &c, const std::valarray<int> &d)
{
    int l = round_length(d);
    std::valarray<int> ind(2);
    double b = 0.0;

    ind[0] = d[0] + max_length;
    ind[1] = d[1] + max_length;
    b += c.encode(rfreq, l);
    b += c.encode_uniform(index[ind[1]][ind[0]], sizes[l]);
    radial_entropy += log((double) sizes[l]) / log(2.0);
    total_bits += b;
    return b;
}

std::valarray<int> radial_coder::decode(decompressor &d)
{
    std::valarray<int> ans(2);
    int l = d.decode(rfreq);
    int inx = d.decode_uniform(sizes[l]);
    ans[0] = rindex[l][inx].x;
    ans[1] = rindex[l][inx].y;
    return ans;
}

double radial_coder::bits() const
{
    return total_bits;
}

double radial_coder::entropy() const
{
    return radial_entropy + rfreq.entropy();
}

std::size_t radial_coder::use() const
{
    return rfreq.use();
}

void radial_coder::dump(std::ostream &out) const
{
    out << "Lengths:\n" << rfreq << "\n";
}

void radial_coder::write_header(std::ostream &out) const
{
    uint32_t ml = max_length;
    out.write((const char *) &ml, sizeof(ml));
}
